      PROGRAM TREEDRIVER_TARG
      IMPLICIT NONE

      INTEGER,PARAMETER :: r8=SELECTED_REAL_KIND(12)

C runtime parameters

      INTEGER :: numpars,farrdim,order,maxparnode,shrink,iflag,numtars
      REAL(KIND=r8) :: theta,dist_tol 

C arrays for coordinates, charge, force calculations 

      REAL(KIND=r8),ALLOCATABLE,DIMENSION(:) :: x,y,z,q
      REAL(KIND=r8),ALLOCATABLE,DIMENSION(:) :: xtar,ytar,ztar
      REAL(KIND=r8),ALLOCATABLE,DIMENSION(:,:) :: tforce
      REAL(KIND=r8),ALLOCATABLE,DIMENSION(:,:) :: dforce

C variables for potential energy computation

      REAL(KIND=r8),ALLOCATABLE,DIMENSION(:) :: tpengtar,dpengtar     

C variables needed for f90 DATE_AND_TIME intrinsic

      INTEGER,DIMENSION(8) :: time1,time2 
      CHARACTER (LEN=8)  :: datec
      CHARACTER (LEN=10) :: timec
      CHARACTER (LEN=5)  :: zonec
      REAL(KIND=r8)      :: totaltime,timedirect,timetree

C variables for error calculations

      REAL(KIND=r8),DIMENSION(3) :: inferr,relinferr,t
      REAL(KIND=r8) :: t1,inferrp,relinferrp

C local variables

      INTEGER :: i,j,err

C EXECUTABLE STATEMENTS

      WRITE(6,*) 'Enter NUMPARS : '
      READ(5,*) numpars
      WRITE(6,*) 'Enter THETA : '
      READ(5,*) theta
      WRITE(6,*) 'Enter ORDER : '
      READ(5,*) order
      WRITE(6,*) 'Enter MAXPARNODE : '
      READ(5,*) maxparnode
      WRITE(6,*) 'Enter SHRINK (0-No 1-Yes) : '
      READ(5,*) shrink
      WRITE(6,*) 'Enter IFLAG  (1-potential energy '
      WRITE(6,*) '             (2-potential energy and force) : '
      READ(5,*) iflag
      WRITE(6,*) 'Enter NUMTARS : '
      READ(5,*) numtars
      WRITE(6,*) 'Enter DIST_TOL : '
      READ(5,*) dist_tol

      ALLOCATE(x(numpars),y(numpars),z(numpars),q(numpars),STAT=err)
      IF (err .NE. 0) THEN
          WRITE(6,*) 'Error allocating arrays for x, y, z and q! '
          STOP
      END IF

      IF (iflag .EQ. 1) THEN
         farrdim=1
      ELSE
         farrdim=numtars
      END IF

      ALLOCATE(tforce(farrdim,3),dforce(farrdim,3),STAT=err)
      IF (err .NE. 0) THEN
          WRITE(6,*) 'Error allocating tforce or tforce! '
          STOP
      END IF

C new arrays for code modification

      ALLOCATE(xtar(numtars),ytar(numtars),ztar(numtars),
     &         tpengtar(numtars),dpengtar(numtars),STAT=err)
      IF (err .NE. 0) THEN
          WRITE(6,*) 'Error allocating arrays for xtar, ytar, ztar, '
          WRITE(6,*) 'tpengtar, and dpengtar! '
          STOP
      END IF

      OPEN(unit=80,file='xyz3d.dat',access='direct',
     *     form='unformatted',recl=24,status='old')
      OPEN(unit=81,file='q3d.dat',access='direct',
     *     form='unformatted',recl=8,status='old')

      DO i=1,numpars
         READ(80,rec=i) x(i),y(i),z(i)
         READ(81,rec=i) q(i)
      END DO
      CLOSE(unit=80)
      CLOSE(unit=81)

C intialiaze targets postitions and charges 

      xtar=x(1:numtars)
      ytar=y(1:numtars)
      ztar=z(1:numtars)

      CALL TREECODE3D_TARG(xtar,ytar,ztar,numtars,dist_tol,tpengtar,
     &                     x,y,z,q,numpars,tforce(1,1),tforce(1,2),
     &                     tforce(1,3),iflag,order,theta,shrink,
     &                     maxparnode,timetree,1,farrdim,numpars)


      IF (iflag .EQ. 1) THEN
         WRITE(6,*) 'Computing potential energy directly'
      ELSE 
         WRITE(6,*) 'Computing potential energy and force directly'
      END IF

      CALL DATE_AND_TIME(datec,timec,zonec,time1)
      CALL DIRECT_ENG_FORCE(xtar,ytar,ztar,numtars,dist_tol,
     &                      dpengtar,x,y,z,q,numpars,iflag,dforce,
     &                      farrdim,numpars)
      CALL DATE_AND_TIME(datec,timec,zonec,time2)
      CALL TTIME(time1,time2,totaltime)
      timedirect=totaltime
      WRITE(6,*) ' '
      WRITE(6,*) 'Direct time (secs) : ',totaltime
      WRITE(6,*) ' '



      inferrp=0.0_r8
      relinferrp=0.0_r8

      DO j=1,numtars
         t1=ABS(tpengtar(j)-dpengtar(j))
         IF (t1 .GT. inferrp) THEN
            inferrp=t1
         END IF
      END DO
      relinferrp=inferrp/MAXVAL(ABS(dpengtar(1:numtars)))

      WRITE(6,*) ' '
      WRITE(6,*) ' '
      WRITE(6,*) 'Absolute and Relative Inf norm error for potential : ' 
      WRITE(6,*) ' '
      WRITE(6,14) inferrp,relinferrp
      WRITE(6,*) ' '

      IF (iflag .NE. 1) THEN

C compute force errors

         inferr=0.0_r8
         relinferr=0.0_r8

         DO j=1,numtars
            DO i=1,3
               t(i)=dforce(j,i)-tforce(j,i)
               IF (ABS(t(i)) .GT. inferr(i)) THEN
                  inferr(i)=ABS(t(i))
               END IF
            END DO
         END DO
         relinferr(1)=inferr(1)/MAXVAL(ABS(dforce(1:numtars,1)))
         relinferr(2)=inferr(2)/MAXVAL(ABS(dforce(1:numtars,2)))
         relinferr(3)=inferr(3)/MAXVAL(ABS(dforce(1:numtars,3)))

C output errors to standard out

         WRITE(6,*) ' '
         WRITE(6,*) 'Absolute Inf norm error for each force component :'      
         WRITE(6,*) ' '
         WRITE(6,13) inferr(1),inferr(2),inferr(3)
         WRITE(6,*) ' '
         WRITE(6,*) 'Relative Inf norm error for each force component :'     
         WRITE(6,*) ' '
         WRITE(6,13) relinferr(1),relinferr(2),relinferr(3)
         WRITE(6,*) ' '

      END IF

 13   FORMAT(E15.8,1X,E15.8,1X,E15.8)
 14   FORMAT(E15.8,1X,E15.8)
      STOP
      END PROGRAM TREEDRIVER_TARG

!!!!!!!!!!!!!!
      SUBROUTINE DIRECT_ENG_FORCE(xtar,ytar,ztar,numtars,dist_tol,
     &                            dpengtar,x,y,z,q,numpars,iflag,
     &                            dforce,farrdim,arrdim)
      IMPLICIT NONE

      INTEGER,PARAMETER :: r8=SELECTED_REAL_KIND(12)
      INTEGER,INTENT(IN) :: farrdim,arrdim,numpars,iflag,numtars 
      REAL(KIND=r8),DIMENSION(arrdim),INTENT(IN) :: x,y,z,q
      REAL(KIND=r8),DIMENSION(numtars),INTENT(IN) :: xtar,ytar,ztar
      REAL(KIND=r8),DIMENSION(numtars),INTENT(INOUT) :: dpengtar
      REAL(KIND=r8),DIMENSION(farrdim,3),INTENT(INOUT) :: dforce
      REAL(KIND=r8),INTENT(IN) :: dist_tol


C local variables 
   
      INTEGER :: i,j
      REAL(KIND=r8) :: tx,ty,tz,dist,xi,yi,zi,qi
      REAL(KIND=r8) :: f1,f2,f3,teng,t1,t2,dist_tolsq

      dist_tolsq=dist_tol*dist_tol

      IF (iflag .NE. 1) THEN

         dforce=0.0_r8
         dpengtar=0.0_r8
         DO i=1,numtars
            xi=xtar(i)
            yi=ytar(i)
            zi=ztar(i)
            f1=0.0_r8
            f2=0.0_r8
            f3=0.0_r8
            teng=0.0_r8
            DO j=1,numpars
               tx=xi-x(j)
               ty=yi-y(j)
               tz=zi-z(j)
               t2=tx*tx+ty*ty+tz*tz
               IF (t2. GT. dist_tolsq) THEN
                  dist=SQRT(t2)
                  t1=1.0_r8/dist
                  dist=t1*t1*t1
                  f1=f1+q(j)*tx*dist
                  f2=f2+q(j)*ty*dist
                  f3=f3+q(j)*tz*dist
                  teng=teng+q(j)*t1
               END IF
            END DO
            dforce(i,1)=f1
            dforce(i,2)=f2
            dforce(i,3)=f3
            dpengtar(i)=teng
         END DO

      ELSE

         dforce=0.0_r8
         dpengtar=0.0_r8
         DO i=1,numtars
            xi=xtar(i)
            yi=ytar(i)
            zi=ztar(i)
            teng=0.0_r8
            DO j=1,numpars
               tx=x(j)-xi
               ty=y(j)-yi
               tz=z(j)-zi
               t2=tx*tx+ty*ty+tz*tz
               IF (t2. GT. dist_tolsq) THEN               
                   teng=teng+q(j)/SQRT(t2)
               END IF
            END DO
            dpengtar(i)=teng
         END DO

      END IF
 
      RETURN
      END SUBROUTINE DIRECT_ENG_FORCE     
